from .rbc_ukqcd_params import set_param, get_param

tag = "a_inv_gev"
set_param("test-4nt16", tag, value = 1.0)
set_param("24D", tag, value = 1.015)
set_param("48I", tag, value = 1.73)
set_param("32IfineH", tag, value = 3.148)

tag = "m_pi"
set_param("test-4nt16", tag, value = 0.2)
set_param("24D", tag, value = 0.13975)
set_param("48I", tag, value = 0.08049)
set_param("32IfineH", tag, value = 0.11790)

tag = "m_kk"
set_param("test-4nt16", tag, value = 0.4)
set_param("24D", tag, value = 0.504154)
set_param("48I", tag, value = 0.28853)
set_param("32IfineH", tag, value = 0.17720)

tag = "zz_vv"
set_param("test-4nt16", tag, value = 0.7)
set_param("24D", tag, value = 0.72672)
set_param("48I", tag, value = 0.71076)
set_param("32IfineH", tag, value = 0.77700)

tag = "zz_aa"
set_param("test-4nt16", tag, value = 0.7)
set_param("24D", tag, value = 0.73457)
set_param("48I", tag, value = 0.71191)
set_param("32IfineH", tag, value = 0.77779)

tag = "m_res"
set_param("test-4nt16", tag, value = 0.001)
set_param("24D", tag, value = 0.0022824)
set_param("48I", tag, value = 0.0006102)
set_param("32IfineH", tag, value = 0.0006296)

tag = "m_l"
set_param("test-4nt16", tag, value = 0.001)
set_param("24D", tag, value = 0.00107)
set_param("48I", tag, value = 0.00078)
set_param("32IfineH", tag, value = 0.0047)

tag = "m_h"
set_param("test-4nt16", tag, value = 0.001)
set_param("24D", tag, value = 0.0850)
set_param("48I", tag, value = 0.0362)
set_param("32IfineH", tag, value = 0.0186)
